package org.fictus

import scala.Array
import net.sf.cglib.proxy.{CallbackFilter, Enhancer, NoOp, Factory}
import java.lang.reflect.{Method, Proxy}
import org.objenesis.ObjenesisHelper
import net.sf.cglib.core.{VisibilityPredicate, CollectionUtils}

object SavedEnhancer {
    private val savedEnhancer = scala.collection.mutable.HashMap.empty[Class[_], Enhancer]

    def put(mockedClass: Class[_], enhancer: Enhancer) = savedEnhancer.put(mockedClass, enhancer)
    def get(mockedClass: Class[_]) = savedEnhancer.get(mockedClass)
}

trait MockProxy {
    def createProxy[A](invoke: Invoke)(implicit context: MockContext, manifest: Manifest[A]) = {
        val mockedClass = manifest.erasure
        val cl = getClass.getClassLoader

        mockedClass.isInterface match {
            case true => Proxy.newProxyInstance(cl, Array(mockedClass), invoke.interfaceInvoke).asInstanceOf[A]
            case false =>
                val enhancer = SavedEnhancer.get(mockedClass) match {
                    case Some(foundEnhancer) => foundEnhancer
                    case None =>
                        val enhancer = createEnhancer(mockedClass)
                        SavedEnhancer.put(mockedClass, enhancer)
                        enhancer
                }
                val mockClass = enhancer.createClass()
                Enhancer.registerCallbacks(mockClass, Array(invoke.classInvoke))
                val factory = ObjenesisHelper.newInstance(mockClass).asInstanceOf[Factory]
                factory.getCallback(0)
                Enhancer.registerCallbacks(mockClass, null)
                factory.asInstanceOf[A]
        }
    }

    private def createEnhancer(toMock: Class[_]) = {
        // Create the mock
        val enhancer = new Enhancer() {
            override def filterConstructors(sc: Class[_], constructors: java.util.List[_]) {
                CollectionUtils.filter(constructors, new VisibilityPredicate(sc, true))
            }
        }
        enhancer.setSuperclass(toMock)
        enhancer.setCallbackType(classOf[net.sf.cglib.proxy.InvocationHandler])
        enhancer
    }
}
